(******************************************************************************
 ******************************************************************************
 ***                                                                        ***
 ***   Set file time for a directory tree (UNICODE!!!) (include file)       ***
 ***   Version [1.10]                               {Last mod 2003-08-07}   ***
 ***                                                                        ***
 ******************************************************************************
 ******************************************************************************

                                 _\\|//_
                                (` * * ')
 ______________________________ooO_(_)_Ooo_____________________________________
 ******************************************************************************
 ******************************************************************************
 ***                                                                        ***
 ***                Copyright (c) 1995 - 2003 by -=Assarbad=-               ***
 ***                                                                        ***
 ***   CONTACT TO THE AUTHOR(S):                                            ***
 ***    ____________________________________                                ***
 ***   |                                    |                               ***
 ***   | -=Assarbad=- aka Oliver            |                               ***
 ***   |____________________________________|                               ***
 ***   |                                    |                               ***
 ***   | Assarbad@gmx.info|.net|.com|.de    |                               ***
 ***   | ICQ: 281645                        |                               ***
 ***   | AIM: nixlosheute                   |                               ***
 ***   |      nixahnungnicht                |                               ***
 ***   | MSN: Assarbad@ePost.de             |                               ***
 ***   | YIM: sherlock_holmes_and_dr_watson |                               ***
 ***   |____________________________________|                               ***
 ***             ___                                                        ***
 ***            /   |                     ||              ||                ***
 ***           / _  |   ________ ___  ____||__    ___   __||                ***
 ***          / /_\ |  / __/ __//   |/  _/|   \  /   | /   |                ***
 ***         / ___  |__\\__\\  / /\ || |  | /\ \/ /\ |/ /\ | DOT NET        ***
 ***        /_/   \_/___/___/ /_____\|_|  |____/_____\\__/\|                ***
 ***       ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~        ***
 ***              [http://assarbad.net | http://assarbad.org]               ***
 ***                                                                        ***
 ***   Notes:                                                               ***
 ***   - my first name is Oliver, you may well use this in your e-mails     ***
 ***   - for questions and/or proposals drop me a mail or instant message   ***
 ***                                                                        ***
 ***~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~***
 ***              May the source be with you, stranger ... ;)               ***
 ***    Snizhok, eto ne tolko fruktovij kefir, snizhok, eto stil zhizn.     ***
 ***                     Vsem Privet iz Germanij                            ***
 ***                                                                        ***
 *** Greets from -=Assarbad=- fly to YOU =)                                 ***
 *** Special greets fly 2 Nico, Casper, SA, Pizza, Navarion, Eugen, Zhenja, ***
 *** Xandros, Melkij, Strelok etc pp.                                       ***
 ***                                                                        ***
 *** Thanks to:                                                             ***
 *** W.A. Mozart, Vivaldi, Beethoven, Poeta Magica, Kurtzweyl, Manowar,     ***
 *** Blind Guardian, Weltenbrand, In Extremo, Wolfsheim, Carl Orff, Zemfira ***
 *** ... most of my work was done with their music in the background ;)     ***
 ***                                                                        ***
 ******************************************************************************
 ******************************************************************************

 LEGAL STUFF:
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

 Copyright (c) 1995-2003, -=Assarbad=- ["copyright holder(s)"]
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
 3. The name(s) of the copyright holder(s) may not be used to endorse or
    promote products derived from this software without specific prior written
    permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                             .oooO     Oooo.
 ____________________________(   )_____(   )___________________________________
                              \ (       ) /
                               \_)     (_/

 ******************************************************************************)

type
  TErrorFunc = function(pw: PWideChar): BOOL; stdcall;

function MyDirCat(dir, filen: PWideChar): PWideChar;
(*
Function
--------
  This function allocates memory large enough to hold a combined string of
  dir+filen and checks wether to append a backslash after dir and before filen.
  RESULTS OF THIS FUNCTION MUST BE FREED USING FreeMem()

Parameters
----------
  dir
    holds a directory name.
  filen
    hold the filename or searchmask ... whatever ... could be anything ;)
*)
var
  k, l: Integer;
  pwc: PWideChar;
begin
// Return nil by default (includes failure)
  result := nil;
// Get directory length in characters (i.e. 1 character is 2 bytes for Unicode)
  l := lstrlenW(dir);
// Check wether we would need to append a backslash
  if dir[l - 1] <> '\' then
// If so reserve one more character
    inc(l);
// Calculate overall size needed
  k := (lstrlenW(filen) + l + 2) * sizeof(WideChar);
// Allocate memory
  GetMem(pwc, k);
// On success ...
  if Assigned(pwc) then
  try
// Fill the allocated buffer with zeroes
    ZeroMemory(pwc, k);
// Append the directory path first
    lstrcatW(pwc, dir);
// Append a backslash if appropriate
    if lstrlenW(dir) < l then
      lstrcatW(pwc, '\');
// Append the filename
    lstrcatW(pwc, filen);
// Return the result ;)
    result := pwc;
  except
  end;
end;

const
  FILE_READ_ATTRIBUTES = $80;
  FILE_WRITE_ATTRIBUTES = $100;

function ActualSetFileTime(fname: PWideChar; dwFileAttributes: DWORD; const ft, la, lw: FILETIME): Boolean;
(*
Function
--------
  Sets the file times for a file. There is not real way to "set last access time"
  because next time you check the filetime by viewing its properties you will
  touch (i.e. access) the file.

Parameters
----------
  fname
    This is the path to the directory which to recurse
  dwFileAttributes
    Current file attributes of the file
  ft
    Is the creation time of the file. See WIN32_FIND_DATA.ftCreationTime
  la
    Is the last access time of the file. See WIN32_FIND_DATA.ftLastAccessTime
  lw
    Is the last write time of the file. See WIN32_FIND_DATA.ftLastWriteTime
*)
var
  hFile: THandle;
begin
  result := False;
// FILE_FLAG_BACKUP_SEMANTICS is there to treat directories and files the same.
// Otherwise you'd get access denied upon trial to open a directory for changing
// its filetime.
  hFile := CreateFileW(fname, FILE_WRITE_ATTRIBUTES, 0, nil, OPEN_ALWAYS, dwFileAttributes or FILE_FLAG_BACKUP_SEMANTICS, 0);
// On success ...
  if hFile <> INVALID_HANDLE_VALUE then
  try
// Set filetimes accordingly
    result := SetFileTime(hFile, @ft, @la, @lw);
  finally
// And close the file
    CloseHandle(hFile);
  end;
end;

procedure SetFTimeZero(var ft: FILETIME);
var
  st: SYSTEMTIME;
begin
  FileTimeToSystemTime(ft, st);
  st.wHour := 0;
  st.wMinute := 0;
  st.wSecond := 0;
  st.wMilliseconds := 0;
  SystemTimeToFileTime(st, ft);
end;

function MySetFileTime(fname: PWideChar; dwFileAttributes: DWORD; const ft, la, lw: FILETIME; smode: Boolean): Boolean;
(*
  All file times have to be given in local time. This function converts them to
  GMT. For further information see comments of ActualSetFileTime().
*)
var
  hFile: THandle;
  _ft, _la, _lw: FILETIME;
begin
// Special mode sets only the file time to 0:00 while leaving the file dates intact!
  if smode then
  begin
    hFile := CreateFileW(fname, FILE_READ_ATTRIBUTES, 0, nil, OPEN_ALWAYS, dwFileAttributes or FILE_FLAG_BACKUP_SEMANTICS, 0);
// On success ...
    if hFile <> INVALID_HANDLE_VALUE then
    try
// Set filetimes accordingly
      if GetFileTime(hFile, @_ft, @_la, @_lw) then
      begin
        SetFTimeZero(_ft);
        SetFTimeZero(_la);
        SetFTimeZero(_lw);
      end
      else
      begin
        _ft := ft;
        _la := la;
        _lw := lw;
      end;
    finally
// And close the file
      CloseHandle(hFile);
    end;
  end
  else
  begin
    _ft := ft;
    _la := la;
    _lw := lw;
  end;
  LocalFiletimeToFileTime(_ft, _ft);
  LocalFiletimeToFileTime(_la, _la);
  LocalFiletimeToFileTime(_lw, _lw);
  result := ActualSetFileTime(fname, dwFileAttributes, _ft, _la, _lw);
end;

procedure RecurseDirAndSetTime(searchmask, directory: PWideChar; const ft, la, lw: FILETIME; smode, allfiles: Boolean; errorfunc: TErrorFunc);
(*
Function
--------
  This funtion call itself recursively to recurse all the subdirectories below
  the directory given in the initial call to RecurseDirAndSetTime().

Parameters
----------
  searchmask
    Can be a combination of wildcards and other characters that forms a search
    expression. E.g.: '*.*'
  directory
    This is the path to the directory which to recurse
    (MUST BE IN THE UNICODE FORM "\\?\C:\Directory\File")
  ft
    Is the creation time of the file. See WIN32_FIND_DATA.ftCreationTime
  la
    Is the last access time of the file. See WIN32_FIND_DATA.ftLastAccessTime
  lw
    Is the last write time of the file. See WIN32_FIND_DATA.ftLastWriteTime
*)
var
  fdw: WIN32_FIND_DATAW;
  hSearch: THandle;
  finished: Boolean;
  dir, fname: PWideChar;
begin
// Just forge together the directory and searchmask ... RESULT MUST BE FREED using FreeMem()
  dir := MyDirCat(directory, searchmask);
// Got the combined directory name
  if Assigned(dir) then
  try
// Zero out the WIN32_FIND_DATA structure!
    ZeroMemory(@fdw, sizeof(fdw));
// Find first file
    hSearch := FindFirstFileW(dir, fdw);
// Only proceed upon success
    finished := hSearch = INVALID_HANDLE_VALUE;
// Success ...
    while not finished do
    begin
// Do not treat the "symbolic" file names "." and ".."
      if ((lstrcmpiW(fdw.cFileName, '.') <> 0) and (lstrcmpiW(fdw.cFileName, '..') <> 0)) then
      begin
// Combine found name and current directory (within recursion)
        fname := MyDirCat(directory, @fdw.cFileName);
// Got it ...
        if Assigned(fname) then
        try
// Set file time and check for success
          if not MySetFileTime(fname, fdw.dwFileAttributes, ft, la, lw, smode) or allfiles then
// If failed then check wether an error function was given
            if Assigned(errorfunc) then
// If so, call it.
              errorfunc(fname);
// If the current found file was a directory ...
          if (fdw.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0 then
// ... let's recurse into it by calling ourself with new directory path
            RecurseDirAndSetTime(searchmask, fname, ft, la, lw, smode, allfiles, errorfunc);
        finally
// Don't forget to free the result of MyDirCat()
          FreeMem(fname);
        end;
      end;
// Zero out the WIN32_FIND_DATA structure!
      ZeroMemory(@fdw, sizeof(fdw));
// Find next file and decide wether to proceed or not
      finished := not FindNextFileW(hSearch, fdw);
    end;
  finally
// Don't forget to free the result of MyDirCat()
    FreeMem(dir);
  end;
end;

